"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const FILTERED = Symbol("filtered AsyncArray element");
const isNotFilteredElement = (elem) => elem !== FILTERED;
const promisifyValue = (val) => Promise.resolve(val);
// Two main characteristics:
// 1. Takes asynchronous functions for operating on arrays.
// 2. Operates on arrays concurrently.
// For example:
// .map maps values to a potential value, and calls the mapping function on elements non-deterministically.
// .some/.every takes asynchronous predicates and races value Promises for first to pass/fail predicate, not necessarily most initial in position.
class AsyncArray {
    constructor(from) {
        this.elements = [...from];
    }
    static from(from) {
        return new AsyncArray(Array.from(from, promisifyValue));
    }
    async toArray() {
        return (await Promise.all(this.elements)).filter(isNotFilteredElement);
    }
    concat(...vals) {
        return new AsyncArray(this.elements.concat(...vals.map((val) => Array.isArray(val)
            ? val.map(promisifyValue)
            : promisifyValue(val))));
    }
    async forEach(iterator) {
        await Promise.all(this.elements.map(async (val) => {
            const resolved = await val;
            if (isNotFilteredElement(resolved)) {
                await iterator(resolved);
            }
        }));
    }
    async every(predicate) {
        // Use .some for short-circuiting behaviour.
        return !(await this.some(async (val) => !(await predicate(val))));
    }
    some(predicate) {
        return new Promise((resolve, reject) => {
            let remaining = this.elements.length;
            this.elements.forEach(async (val) => {
                try {
                    const resolved = await val;
                    if (isNotFilteredElement(resolved) && (await predicate(resolved))) {
                        resolve(true);
                    }
                }
                catch (e) {
                    reject(e);
                }
                if (--remaining == 0) {
                    resolve(false);
                }
            });
        });
    }
    map(mapper) {
        return this.next((val) => promisifyValue(mapper(val)));
    }
    filter(predicate) {
        return this.next(async (elem) => (await predicate(elem)) ? elem : FILTERED);
    }
    // TypeScript does not allow asynchronous user-defined type guards currently.
    filterType(predicate) {
        return this.filter(predicate);
    }
    // Hack to allow filtering using a synchronous user-defined type guard.
    push(...vals) {
        return this.elements.push(...vals.map(promisifyValue));
    }
    async pop() {
        let popped;
        while ((popped = await this.elements.pop()) === FILTERED) { }
        return popped;
    }
    unshift(vals) {
        return this.elements.unshift(...vals.map(promisifyValue));
    }
    async shift() {
        let shifted;
        while ((shifted = await this.elements.shift()) === FILTERED) { }
        return shifted;
    }
    next(fn) {
        return new AsyncArray(this.elements.map(async (val) => {
            const resolved = await val;
            return resolved !== FILTERED ? fn(resolved) : FILTERED;
        }));
    }
}
exports.default = AsyncArray;
